;; Build and maintain an irc connection.
;; Copyright (C) 2001  Neal H Walfield  <neal@cs.uml.edu>
;;
;; This file is part of the Deride
;;
;; Deride is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;; 
;; Deride is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with the GNU Hurd; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(define-module (irc))

(use-modules (ice-9 slib))
(use-modules (ice-9 regex))
(use-modules (rfc-2812))

;; The End of line for irc servers.
(define-public irc-eol "\n\r")

;; Connect to SERVER on PORT.  Return a SOCKET.
(define-public (irc-connect server port)
  (let*
      ((s (socket AF_INET SOCK_STREAM 0))
       (host (catch
	      #t
	      (lambda () (hostent:addr-list (gethostbyname server)))
	      (lambda args #f))))

    (and host
	 (connect s AF_INET (car host) port)
	 s)))

;; Send DATA to SOCKET and the STDOUT
(define-public (irc-send-raw data socket)
  (display (string-append data "\n"))
  (write-line (string-append data irc-eol) socket))

;; Login to the irc server on SOCKET using NICK and REALNAME
(define-public (irc-send-login socket nick realname)
  (irc-send-raw (string-append "USER " nick " 0 * :" realname)
		       socket)
  (irc-send-raw (string-append "NICK " nick) socket))

;; Outstanding number of pings to the server
(define irc-outstanding-server-pings 0)

;; Ping the server.
(define (irc-ping-server socket)
  (irc-send-raw (string-append "PING " irc-server-name) socket)
  (set! irc-outstanding-server-pings (1+ irc-outstanding-server-pings)))

;; The name of the irc server we are directly connected to (as it sees
;; fit to call itself).
(define irc-server-name "unknown")

;; Parse a packet and act on it!
(define (parse-packet data socket)
  ;; Parse an irc message RFC1459, section 2.3.1
  (define irc-message-regex
    (make-regexp
     (string-append
      "^(:(([^ !]+)!?([^ ]*)))? *"      ; The prefix: i.e. `:nick!user@host'
      "([a-zA-Z]+|[0-9]{3}) +"          ; The command
      "(([^: ]*) )?"                    ; The first middle parameter
      "(([^:]*):?)?"                    ; The rest of the middle parameters
      "(.*)[\n\r]+?$")))                ; Trailing

  (let ((rx (regexp-exec irc-message-regex data)))
    (if (not rx)
	(client-display (string-append "Unparsable message from server: `"
				       data "'\n"))
	(let ((prefix (or (match:substring rx 2) ""))
	      (sender (or (match:substring rx 3) ""))
	      (senders-host (or (match:substring rx 4) ""))
	      (command (or (match:substring rx 5) ""))
	      (first-middle (or (match:substring rx 7) ""))
	      (middle (or (match:substring rx 9) ""))
	      (trailing (or (match:substring rx 10) "")))

	  (cond

	   ;; Message.
	   ;; By far the most common message.  Ergo, it is at the top.
	   ((string-ci=? command "PRIVMSG")
	    (client-display (string-append "<" sender "/" first-middle "> "
					   trailing "\n")))

	   ;; Notices.  The same as PRIVMSG, however, we are prohibited from
	   ;; replying to them.
	   ((string-ci=? command "NOTICE")
	    (client-display
	     (string-append
	      (if (string=? prefix "")
		  (string-append "{" irc-server-name "/" first-middle "}")
		  (string-append "<" sender "/" first-middle ">"))
	      " " trailing "\n")))

	   ;; PING!
	   ((string-ci=? command "PING")
	    (irc-send-raw (string-append
			   "PONG " (substring data 6))
			  socket))

	   ;; PONG!
	   ((string-ci=? command "PONG")
	    (if (string-ci=? prefix irc-server-name)
		(set! irc-outstanding-server-pings
		      (1- irc-outstanding-server-pings))
		(client-display (string-append "PONG from " prefix "\n"))))

	   ;; Welcome to the irc server, blah, blah, blah.  Save the
	   ;; server's name so we can ping it later; this is the easiest
	   ;; spot to get it.
	   ((or
	     (string=? command RPL_WELCOME)
	     (string=? command RPL_YOURHOST)
	     (string=? command RPL_CREATED)
	     (string=? command RPL_MYINFO))
	    (set! irc-server-name prefix)
	    (client-display (string-append "[info] " trailing "\n")))

	   ;; Server info; just echo it.
	   ((or
	     (string=? command RPL_LUSERCLIENT)
	     (string=? command RPL_LUSEROP)
	     (string=? command RPL_LUSERUNKNOWN)
	     (string=? command RPL_LUSERCHANNELS)
	     (string=? command RPL_LUSERME))
	    (client-display (string-append "[info] " trailing "\n")))

	   ;; Try to connect to a different server.
	   ((string=? command RPL_BOUNCE)
	    (client-display (string-append "Being redirected!\n"
					   trailing "\n")))

	   ;; Got the topic
	   ;; "<channel> <mode> <mode params>"
	   ((string=? command RPL_CHANNELMODEIS)
	    (display "modes\n"))

	   ;; "<channel> :No topic is set"
	   ((string=? command RPL_NOTOPIC)
	    (display "No topic for foo\n"))

	   ;; "<channel> :<topic>"
	   ((string=? command RPL_TOPIC)
	    (display "Topic\n"))

	   ;; MOTD
	   ((or (string=? command RPL_MOTDSTART)
		(string=? command RPL_MOTD)
		(string=? command RPL_ENDOFMOTD))
	    (client-display (string-append "[motd] " trailing "\n")))

	   ((client-display (string-append "Unknown command: `"
					   prefix "' `"
					   command "' `"
					   first-middle "' `"
					   middle "' `"
					   trailing "'\n"))))))))

;; Enter the server loop.  Read a line from SOCKET and pass it to
;; parse-packet.  If parse-packet returns false or EOF of file on SOCKET,
;; is reached, return.  Otherwise, loop.
(define-public (irc-loop socket)
  (if (eqv? (car (select (list (port->fdes socket)) '() '() 200))
	    '())
      (if (> irc-outstanding-server-pings 0)
	  (begin (display "Ping timeout\n")
		 (if (not (port-closed? socket))
		     close-port socket))
	  (and (irc-ping-server socket)
	       (irc-loop parse-packet socket)))
      (let ((data (read-line socket)))
	(or (eof-object? data)
	    (not (parse-packet (substring data 0 (1- (string-length data)))
			    socket))
	    (irc-loop socket)))))
